#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define pi 3.14159265

#include "./samc.h"

/***********************************************************************************************
 NOTES:
  Input data set should contain an intercept column if an intercept is desired when fitting.
  The file of groups consists of 2 columns: 
    Column 1 is the group number labeled 1, 2, ..., N.
    Column 2 is the comma seperated list of column numbers.
  The file of interactions consists of 3 columns of group numbers: 
    Column 1 is the group number for the interaction.
    Column 2 is the group number for one of the main efects.
    Column 3 is the group number for the other main effect.

  The vector indx gives the original ordering for the new groups, so that indx[i] is the original
      group id of the ordered ith group.
  The vector ranks (inverse of indx) gives the rankings for the original group ids, so that ranks[i]
      is the ranking (new group id) for the original ith group.
**************************************************************************************************/

static int resample_main(argc, argv)
int argc;
char *argv[];
{
  int stime;
  long ltime;
  int *x,*y,*truex,*besty,*indx,dim,tail,i,j,k,iter,grid, ii;
  double **hist,*mar,*mary,logw, totalweight,delta,fvalue;
  double sum,ss,min,max,ave,un,var,*fx,*pw, *model;
  double deviance, skewness,z;
  FILE *ins=NULL, *ins2=NULL, *fpt_prob=NULL, *fpt_log=NULL, *fpt_hist=NULL, *fpt_best=NULL;
  RET_COST *rcost, **Models, *rcost_best;
  char *tempstr, sep;
  PARMS *parms;
  int nc, nr, *ivec, header, **tempX, interFlag, grp, **tmpMat, checkInt, histFlag;
  double rho, scale, lowE, maxE, **DataX, *DataY, *refden, *cor, *fsave, bestvalue, accept_loc, total_loc;
  int stepscale, savestep, warm, total_iteration, DataNum, DataP, TotalNcol, Data0X_n, maxvar, *Data0X;
  int **GroupX, *GroupN, Sum_GroupN, **IntX, IntX_flag, sze, *bestx, **modsave, *ranks, *IntX_N, *ivec2;

  /* Get the parameters */
  parms = (PARMS *) malloc(sizeof(PARMS));
  CHECK_MEM(parms);

  /* Initialize some parms before parm file is read */
  parms->Model_maxn = MODEL_MAXN;
  parms->seed       = -1;
  parms->maxbeta    = MAX_BETA;

  /* Read the parameter file */
  read_parms(parms, argv[1]);

  /* Read in the hist file to determine options lowE, maxE, and scale.
     rho is always 1, stepscale is 100 */
  histFlag = 9;
  grid     = 0;
  hist     = set_hist(argv[1], &grid, &histFlag, parms);
  if (histFlag) {
    if (hist) {
      free_dmatrix(hist,0,grid,1,4);
      hist = NULL;
    }
    return(-2);
  }
  parms->rho       = 1.0;
  parms->lowE      = hist[0][1];
  parms->maxE      = hist[grid][1];
  parms->scale     = 1.0/(hist[1][1] - hist[0][1]);
  parms->stepscale = 100;

/*
printf("lowE=%g, maxE=%g, scale=%g, grid=%d\n", parms->lowE, parms->maxE, parms->scale, grid);
*/

  total_iteration = parms->total_iteration;
  warm            = parms->warm;
  savestep        = parms->savestep;
  lowE            = parms->lowE;
  maxE            = parms->maxE;
  scale           = parms->scale;
  rho             = parms->rho;
  stepscale       = parms->stepscale;
  checkInt        = parms->check_interactions;

  /* Set up the models structure */
  set_Models(argv[1], parms);
  Models = parms->Models;
  
  /* Allocate memory for temporary character strings */
  tempstr = (char *) malloc(MAXFILELEN*sizeof(char));
  CHECK_MEM(tempstr);

  /* Get the name of the input file */
  get_option(argv[1], "Data", tempstr, 1);

  /* Get the file delimiter */
  sep = getDelimiter(tempstr, BUFFERSIZE);

  /* Get the number of rows and columns */
  nr = getNrows(tempstr, BUFFERSIZE);
  nc = getNcols(tempstr, sep, BUFFERSIZE);

  /* Determine if the file has a header of variable names */
  header = getHeader(tempstr, sep, nc, BUFFERSIZE);
  if (header) nr--;
  DataNum = nr;
  parms->DataNum = DataNum;
  TotalNcol = nc;
  parms->TotalNcol = TotalNcol;
 
  /*
  printf("Nrows=%d, Ncols=%d, header=%d\n", nr, nc, header);
  */

  /* Read the data into a matrix */
  DataX = dmatrix(1, nr, 1, nc); 
  parms->DataX = DataX;
  read_data(tempstr, sep, DataX, nr, nc, BUFFERSIZE, header);

  /*
  print_dMat(DataX, 1, 10, 1, MIN(10, nc), "Subset of data file:");
  */

  /* Store the response */
  DataY = dvector(1, DataNum);
  parms->DataY = DataY;
  k = parms->response_col;
  for (i=1; i<=DataNum; i++) DataY[i] = DataX[i][k];

  /* Get the columns to be always included */
  get_option(argv[1], "Data0_cols", tempstr, 1);
  ivec = ivector(0, nc+1);
  k = getIntsFromStr(tempstr, 10, nc+1, ivec);

  /* Get the number of cols to always include */
  for (i=1, k=0; i<=nc; i++) k += ivec[i];
  Data0X_n = k;
  parms->Data0X_n = Data0X_n;

  /* Put the column number always to include in Data0X vector */
  if (Data0X_n) {
    Data0X = ivector(1, Data0X_n);
    for (i=1, k=0; i<=nc; i++) {
      if (ivec[i]) {
        k++;
        Data0X[k] = i;    
      }
    }
  } else {
    Data0X = NULL;
  }
  parms->Data0X = Data0X;

  /* Define the groups. Groups must be defined */
  ivec[parms->response_col] = 1;
  get_option(argv[1], "groups", tempstr, 1);

  tempX = imatrix(1, nc, 1, nc);
    
  /* Read in the file of groups and set up GroupX and GroupN */
  DataP = read_groups(tempstr, ivec, nc, tempX);
  parms->DataP = DataP;
  maxvar = MIN(parms->maxvar, DataP);
  parms->maxvar = maxvar;

  GroupX = imatrix(1, DataP, 1, DataP);
  GroupN = ivector(1, DataP);
  Sum_GroupN = 0;
  for (i=1, grp=0; i<=nc; i++) {
    /* Get the number of columns in this group */
    for (j=1, k=0; j<=nc; j++) k += tempX[i][j];
    if (k) {
      grp++;
      GroupN[grp] = k;
      Sum_GroupN += k;
      for (j=1, k=0; j<=nc; j++) {
        if (tempX[i][j]) {
          k++;
          GroupX[grp][k] = j;
        }
      }
    }
  }
  parms->GroupX = GroupX;
  parms->GroupN = GroupN;
  parms->Sum_GroupN = Sum_GroupN;

  if (tempX) free_imatrix(tempX,1,nc,1,nc);
  free_ivector(ivec, 0, nc+1);

  /* Set up the interaction matrix, not using this feature at this moment */
  interFlag = 1 - get_option(argv[1], "interactions", tempstr, 0);

  if (interFlag) {
    /* The interaction data file can have a different number of columns for each row.
       Column 1 is the interaction group number, followed by the other group numbers that must be in the model.
    */
    IntX_N = ivector(1, DataP);
    IntX = read_interactions(tempstr, DataP, IntX_N);
    IntX_flag = 1;
  } else {
    IntX_flag = 0;
    IntX = NULL;
    IntX_N = NULL;
  }
  parms->IntX_flag = IntX_flag;
  parms->IntX = IntX;
  parms->IntX_N = IntX_N;

  /* Open the output files */
  i = get_option(argv[1], "out1", tempstr, 0);
  if (!i) ins = open_file(tempstr, "w");
  i = get_option(argv[1], "out2", tempstr, 0);
  if (!i) ins2 = open_file(tempstr, "w");

  i = get_option(argv[1], "out_prob", tempstr, 0);
  if (!i) fpt_prob = open_file(tempstr, "w");

  /*
  i = get_option(argv[1], "out_log", tempstr, 0);
  if (!i) fpt_log = open_file(tempstr, "w");
  */
  /*
  i = get_option(argv[1], "out_hist", tempstr, 0);
  if (!i) fpt_hist = open_file(tempstr, "w");
  */
  i = get_option(argv[1], "out_best", tempstr, 0);
  if (!i) fpt_best = open_file(tempstr, "w");

  free(tempstr);

  /* DataNum tell the allocatoin function to allocate memory for yhat */
  rcost = cost_init(DataNum, DataP, TotalNcol);
  /*rcost_best = cost_init(DataNum, DataP, TotalNcol); */

  /* initialize the random number generator */
  if (parms->seed > -1) {
    srand(parms->seed);
    if (fpt_log) {
      fprintf(fpt_log, "random seed=%d \n", parms->seed);
      fflush(fpt_log);
    }
  } else {
    ltime=time(NULL);
    stime=(unsigned int)ltime/2;
    srand(stime);
    if (fpt_log) {
      fprintf(fpt_log, "random seed=%d \n", stime);
      fflush(fpt_log);
    }
  }
  if (fpt_log) fclose(fpt_log);

  /* Set up the hist matrix. hist matrix starts at 0 */
  /*
  grid = ceil((maxE-lowE)*scale);
  histFlag = 9;
  hist = set_hist(argv[1], &grid, &histFlag, parms);
  if (histFlag) {
    if (hist) {
      free_dmatrix(hist,0,grid,1,4);
      hist = NULL;
    }
    return(-2);
  }
  */

  sze  = grid;
  parms->sze = sze;

  refden=dvector(0,grid);
  parms->refden = refden;

  x=ivector(1,DataP);
  y=ivector(1,DataP);
  truex=ivector(1,DataP);
  bestx=ivector(1,DataP);
  parms->bestx = bestx;
  besty=ivector(1,DataP);
  /*DataZ=dmatrix(1,DataNum,1,DataP);*/
  mar=dvector(1,DataP);
  mary=dvector(1,DataP);
  pw=dvector(1,DataP);
  fx=dvector(1,DataP);
  indx=ivector(1,DataP);
  parms->indx = indx;
  cor=dvector(1,DataP);
  parms->cor = cor;
 
  /* the 0 position is for the model using only offset */
  modsave=imatrix(0,maxvar,1,DataP);
  parms->modsave = modsave;
  fsave=dvector(0,maxvar); 
  parms->fsave = fsave;
  model = dvector(0, maxvar);
  for (i=0; i<=maxvar; i++) model[i] = 0.0;


  /*PenaltyC=log(1.0*DataNum)+2.0*gamma*log(1.0*DataP);*/
  /*  PenaltyC=log(1.0*DataNum)+2.0*gamma*log(1.0*DataP); */

  ranks = ivector(1, DataP);
  for (i=1; i<=DataP; i++) {
    ranks[i] = i;
    indx[i]  = i;
  }
  parms->ranks = ranks;

  /* start from here */
 
  /* rank individual variable */
  if (DEBUG) printf("Begin: Rank variables\n");
  min=1.0e+100; 
  max=-1.0e+100;
  for(sum=var=0.0, i=1; i<=DataP; i++){
     for(j=1; j<=DataP; j++) x[j]=0;
     x[i]=1; /* changed */
     cost(x, rcost, 0, parms);  /* Do not check interactions here */
     fx[i] = rcost->value;

     if(fx[i]<1.0e+100){ sum+=fx[i]; var+=fx[i]*fx[i]; }
     if(fx[i]<min) min=fx[i];
       else if(fx[i]>max) max=fx[i];

    }
  ave=sum/DataP; 
  var=(var - DataP*ave*ave)/(DataP-1);
  for(i=1; i<=DataP; i++)
    {
       pw[i]=-1.0*exp(-(fx[i]-min)/(5.0*sqrt(var))-0.1);
       if(pw[i]> -0.01) pw[i]=-0.01;
    }
 
  if (DEBUG) printf("END: Rank variables\n");  

  /* indx gives the ordering for the groups, DataP is ordered*/
  indexx(DataP, pw, indx);

  /* Reorder x and change the values in GroupX, GroupN, and intX */
  tempX  = imatrix(1, DataP, 1, DataP);
  tmpMat = imatrix(1, DataP, 1, DataP);
  ivec   = ivector(1, DataP);
  ivec2  = ivector(1, DataP);
  for(i=1; i<=DataP; i++) ivec2[i] = 0;
  for(i=1; i<=DataP; i++){
    k = indx[i];
    ranks[k] = i;
    ivec[i] = GroupN[k];
    for (j=1; j<=DataP; j++) tempX[i][j] = GroupX[k][j];
    if (IntX_flag) {
      for (j=1; j<=IntX_N[k]; j++) tmpMat[i][j] = IntX[k][j-1];
      ivec2[i] = IntX_N[k];
    }
  }
  for(i=1; i<=DataP; i++){
    GroupN[i] = ivec[i];
    for (j=1; j<=DataP; j++) GroupX[i][j] = tempX[i][j];
    if (IntX_flag) {
      k = ivec2[i];
      /* Free memory, then re-allocate */
      if (IntX[i]) free(IntX[i]);
      IntX[i] = (int *) malloc(k*sizeof(int));
      CHECK_MEM(IntX[i]);
      for (j=1; j<=k; j++) IntX[i][j-1] = tmpMat[i][j];
      IntX_N[i] = k;
    }
  }
  free_ivector(ivec, 1, DataP);
  free_ivector(ivec2, 1, DataP);
  free_imatrix(tempX, 1, DataP, 1, DataP);
  free_imatrix(tmpMat, 1, DataP, 1, DataP);
 
  /* cor is the ordered pw */
  for(j=1; j<=DataP; j++){
      cor[j]=-1.0*pw[indx[j]];
  }

  /* Initialization */
  for(sum=0.0, ss=1.0, i=0; i<=sze; i++){ ss+=i; refden[i]=1.0; sum+=refden[i]; }
  for(i=0; i<=sze; i++) refden[i]/=sum;


  /*so the last row is maxE+1.0/scale */    
  /* ---hist[0]---hist[1]---hist[2]---max---max+1 */

    /* initialization of samx */
  for(i=1; i<=DataP; i++)
    { 
      x[i]=0;
    }

  /* select from the most likely one */
  j = model_max_value(parms);
  ii = Models[j]->num_grp;
  if (ii > 0)
    { 
      for(i=1; i<=ii; i++)
	{ /* Get the new group ids, since the Models structure saves the original group ids */
	  x[ranks[Models[j]->groups[i]]] = 1;
	}
    }
  

  fvalue = Models[j]->value;

  deviance = Models[j]->deviance;
  bestvalue=fvalue;
  parms->bestvalue = bestvalue;

  /* Structure to save the best model */
  rcost_best = cost_init(0, DataP, TotalNcol);
  cost_copy(Models[j], 0, rcost_best);

  /* dim should be the number of columns in x */
  for(i=1; i<=DataP; i++)
  { 
     bestx[i]=x[i]; 
  }

    
  accept_loc=total_loc=0.1; totalweight=0.0; skewness=0.0;
  parms->accept_loc = accept_loc;
  parms->total_loc  = total_loc;

   /**********************************************************************/
   
   for(i=0; i<=maxvar; i++) fsave[i]=1.0e+200;
   max=-1.0e+100;
   /* hist[,4] is the weight, the 2nd column is the log weight */
   for(i=0; i<=sze; i++){
     un = exp(hist[i][2]);
     if(un>max) max=un;
   }

   iter = 0;
   ii = 0;
   for (iter=1; iter <= (total_iteration+warm); iter++)
     {
       delta = 0.0;
       metmove(x,&fvalue,hist,delta,&logw,rcost, checkInt, 0, parms); 

       /* See if this model is better */
       if (rcost->value < rcost_best->value) cost_copy(rcost, 0, rcost_best);

       if ((iter > warm) && (iter%savestep==0))
        { 
	  dim = rcost->dim;
	  deviance=rcost->deviance;
	  un=0.0;
	  while(un<=0.0 || un>=1.0) un=rand()*1.0/RAND_MAX;
	  

	  if(un<exp(logw)/max)
	    {  
	      ii = ii + 1;
	      if(deviance<0.2) deviance=rand()*0.1/RAND_MAX+0.1;
	      un=gammpln(0.5*(DataNum-dim),0.5*deviance,&tail);
	      z=tail*inverse_normal_cdf_log(un);
	    
	      indexx_convert_integer(DataP,indx,x,y);    

	      if (ins) {
		fprintf(ins, " %d %d %g %g %g %g %g \n", dim, iter, deviance,
			fvalue, logw, z, exp(logw)/max); 
		fflush(ins);
	      }
	      if (ins2) {
		fprintf(ins2, "%d\n", dim);
		for (j=1; j<dim; j++) fprintf(ins2, "%d ", (rcost->col_ids)[j]);
              fprintf(ins2, "%d\n", (rcost->col_ids)[dim]);
		
		for (j=1; j<dim; j++) fprintf(ins2, "%g ", (rcost->beta)[j]);
              fprintf(ins2, "%g\n", (rcost->beta)[dim]);
		
		fflush(ins2);
	      }

           
	    } /* end of if (un..) */

	}
       
       /*       printf("iter=%d, ii = %d  \n",iter, ii);*/
       


   } /* END: for(iter=1; iter<=warm+total_iteration; iter++) */
   if (ins2) fclose(ins2); 
   if (ins) fclose(ins); 
  
  /* Write out best model */
  if (fpt_best) {
    write_model(fpt_best, rcost_best, 1);
    fclose(fpt_best);
  }
  
  if (fpt_hist) fclose(fpt_hist);
  if (fpt_prob) fclose(fpt_prob);
 
  free_dmatrix(hist,0,grid,1,4);
  free_dvector(refden,0,grid);
  free_ivector(x,1,DataP);
  free_ivector(y,1,DataP);
  free_ivector(truex,1,DataP);
  free_ivector(bestx,1,DataP);
  free_ivector(besty,1,DataP);
  if (DataX != NULL) free_dmatrix(DataX,1,DataNum,1,TotalNcol);
  if (Data0X != NULL) free_ivector(Data0X,1,Data0X_n);
  /*free_dmatrix(DataZ,1,DataNum,1,DataP);*/
  free_dvector(DataY,1,DataNum);
  free_dvector(mar,1,DataP);
  free_dvector(mary,1,DataP);
  free_dvector(pw,1,DataP);
  free_dvector(fx,1,DataP);
  free_ivector(indx,1,DataP);
  free_dvector(cor,1,DataP);
  free_imatrix(modsave,0,maxvar,1,DataP);
  free_dvector(fsave,0,maxvar);
  free_dvector(model,0,maxvar);
  free_ivector(GroupN,1,DataP);
  free_imatrix(GroupX,1,DataP,1,DataP);
  free_ivector(ranks, 1, DataP);

  rcost->num_grp = DataP;
  rcost->dim     = TotalNcol;
  cost_free(rcost, DataNum);

  rcost_best->num_grp = DataP;
  rcost_best->dim     = TotalNcol;
  cost_free(rcost_best, DataNum);

  if (parms->IntX_N) free_ivector(parms->IntX_N, 1, DataP);
  if (parms->IntX) matrix_free((void **) parms->IntX, DataP+1);

  models_free(parms);
  free(parms);

  return(0);

} /* END: resample_main */


/* Function that R will call */
void resample(pargc, argv, ret)
int *pargc, *ret;
char **argv;
{
  *ret = 1;
  *ret = resample_main(*pargc, argv);

  return;

} /* END: resample */
